/*
 ************************************************************************
 *                                                                      *
 * udpBezier -- udp file to generate a bezier wire, sheet, or solid body*
 *                                                                      *
 *            Written by Bridget Dixon @ Syracuse University            *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                      John Dannenhoffer @ Syracuse University         *
 *                                                                      *
 ************************************************************************
 */

#ifdef GEOM_EGADS

/*
 * Copyright (C) 2013  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef WIN32
    #define snprintf _snprintf
#endif

#define CHECK_STATUS(X)                                           \
    if (status < EGADS_SUCCESS) {                                 \
        printf( "ERROR:: BAD STATUS = %d from %s\n", status, #X); \
        goto cleanup;                                             \
    } else if (Debug[0]) {                                        \
        printf("%s -> status=%d\n", #X, status);                  \
    }

#include "egads.h"

/* ADDITIONAL ATTRIBUTE TYPE */
#define ATTRREALSEN 4

#define TWOPI 6.2831853071795862319959269

/* static variables used to hold info between function calls */
static char   *FileName = NULL;

/* number of Bezier UDPs */
static int    numUdp    = 0;

/* arrays to hold info for each of the UDPs */
static ego    *ebodys   = NULL;
static int    *Debug    = NULL;
static int    *Imax     = NULL;
static int    *Jmax     = NULL;
static double **Cp      = NULL;


/*
 ************************************************************************
 *                                                                      *
 *   udpErrorStr - print error string                                   *
 *                                                                      *
 ************************************************************************
 */

static /*@null@*/ char *
udpErrorStr(int stat)
{
    char *string;

    string = EG_alloc(25*sizeof(char));
    if (string == NULL) {
        return string;
    }
    snprintf(string, 25, "EGADS status=%d", stat);

    return string;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpInitialize - initialize and get info about the list of arguments*
 *                                                                      *
 ************************************************************************
 */

int
udpInitialize(int    *nArgs,
              char   ***namex,
              int    **typex,
              int    **idefaulx,
              double **ddefaulx)
{
    int    *type, *idefault;
    char   **name;
    double *ddefault;

    /* make the arrays */
    ebodys = (ego    *) EG_alloc(sizeof(ego    ));
    Debug  = (int    *) EG_alloc(sizeof(int    ));
    Imax   = (int    *) EG_alloc(sizeof(int    ));
    Jmax   = (int    *) EG_alloc(sizeof(int    ));
    Cp     = (double**) EG_alloc(sizeof(double*));

    if (ebodys == NULL || Debug == NULL ||
        Imax   == NULL || Jmax  == NULL || Cp == NULL) {
        return EGADS_MALLOC;
    }

    /* initialize the array elements that will hold the "current" settings */
    ebodys[0] = NULL;
    Debug[ 0] = 0;
    Imax[  0] = 1;
    Jmax[  0] = 1;
    Cp[    0] = NULL;

    /* set up returns that describe the UDP */
    *nArgs    = 2;
    *namex    = NULL;
    *typex    = NULL;
    *idefaulx = NULL;
    *ddefaulx = NULL;

    *namex = name = (char **) EG_alloc((*nArgs)*sizeof(char *));
    if (*namex == NULL) {
        return EGADS_MALLOC;
    }

    *typex = type = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*typex == NULL) {
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *idefaulx = idefault = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*idefaulx == NULL) {
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *ddefaulx = ddefault = (double *) EG_alloc((*nArgs)*sizeof(double));
    if (*ddefaulx == NULL) {
        EG_free(*idefaulx);
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    name[0]     = "FileName";
    type[0]     = ATTRSTRING;
    idefault[0] = 0;
    ddefault[0] = 0.0;

    name[1]     = "Debug";
    type[1]     = ATTRINT;
    idefault[1] = 0;
    ddefault[1] = 0.0;

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpReset - reset the arguments to their defaults                   *
 *                                                                      *
 ************************************************************************
 */

int
udpReset(int flag)
{
    int   iudp;

    if (FileName != NULL) {
        EG_free(FileName);
        FileName = NULL;
    }

    /* reset the "current" settings */
    if (flag == 0) {
        Debug[0] = 0;
        Imax[ 0] = 1;
        Jmax[ 0] = 1;

        if (Cp[0] != NULL) {
            EG_free(Cp[0]);
            Cp[0] = NULL;
        }

    /* called when closing up */
    } else {
        for (iudp = 0; iudp <= numUdp; iudp++) {
            if (Cp[iudp] != NULL) {
                EG_free(Cp[iudp]);
                Cp[iudp] = NULL;
            }
        }

        EG_free(ebodys); ebodys = NULL;
        EG_free(Debug );  Debug = NULL;
        EG_free(Imax  );  Imax  = NULL;
        EG_free(Jmax  );  Jmax  = NULL;
        EG_free(Cp    );  *Cp   = NULL;

        numUdp = 0;
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSet - set an argument                                           *
 *                                                                      *
 ************************************************************************
 */

int
udpSet(char *name,
       char *value)
{
    int  i, len;
    char *pEnd, lowername[257];

    if        (name  == NULL) {
        return EGADS_NONAME;
    } else if (strlen(name) == 0 || strlen(name) > 255) {
        return EGADS_NONAME;
    } else if (value == NULL) {
        return EGADS_NULLOBJ;
    }

    for (i = 0; i <= strlen(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    len = strlen(value);
    if (len == 0) {
        return EGADS_NODATA;
    }

    if        (strcmp(lowername, "filename") == 0) {
        if (FileName != NULL) {
            EG_free(FileName);
            FileName = NULL;
        }

        FileName = (char *) EG_alloc((len+1)*sizeof(char));
        if (FileName == NULL) {
            return EGADS_MALLOC;
        }

        strcpy(FileName, value);

    } else if (strcmp(lowername, "debug") == 0) {
        Debug[0] = strtol(value, &pEnd, 10);

    } else {
        printf(" udpSet: Parameter \"%s\" not known (should be filename)\n", name);
        return EGADS_INDEXERR;

    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,
           ego  *ebody,
           int  *nMesh,
           char **string)
{
    int     status = EGADS_SUCCESS;

    int     i, j, imax, jmax, ipat, nipat, jpat, njpat, header[5], periodic;
    int     i0, i1, j0, j1, idum;
    int     inode, iedge, jedge, iface, icp, icp0, icp1, *senses=NULL, *esense=NULL;
    int     oclass, mtype, nchild, *senses2, count, newnode, newedge;
    double  cp[48], trange[4], data[18], tol=1.0e-7;
    FILE    *fp;
    ego     ecurve, esurf, eloop, eshell, etemp[8], eref, *echild;
    ego     *enodes=NULL, *eedges=NULL, *efaces=NULL;

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* increment number of UDPs */
    numUdp++;

    /* increase the arrays to make room for the new UDP */
    ebodys = (ego    *) EG_reall(ebodys, (numUdp+1)*sizeof(ego    ));
    Debug  = (int    *) EG_reall(Debug,  (numUdp+1)*sizeof(int    ));
    Imax   = (int    *) EG_reall(Imax,   (numUdp+1)*sizeof(int    ));
    Jmax   = (int    *) EG_reall(Jmax,   (numUdp+1)*sizeof(int    ));
    Cp     = (double**) EG_reall(Cp,     (numUdp+1)*sizeof(double*));

    if (ebodys == NULL || Debug == NULL || 
        Imax   == NULL || Jmax  == NULL || Cp == NULL) {
        return EGADS_MALLOC;
    }

    ebodys[numUdp] = NULL;
    Debug[ numUdp] = Debug[0];

    /* check that filename was given */
    if (FileName == NULL) {
        status = EGADS_NODATA;
        goto cleanup;
    }

    /* open the file */
    fp = fopen(FileName, "r");
    if (fp == NULL) {
        status = EGADS_NOTFOUND;
        goto cleanup;
    }

    /* read the size of the bezier */
    count = fscanf(fp, "%d %d", &imax, &jmax);
    if (count != 2) {
        status = EGADS_NODATA;
        goto cleanup;
    }

    if        (imax < 4 || (imax-1)%3 != 0) {
        printf(" udpExecute: bad value for imax=%d (from file)\n", imax);
        status = EGADS_NODATA;
        goto cleanup;
    } else if (jmax < 1 || (jmax-1)%3 != 0) {
        printf(" udpExecute: bad value for jmax=%d (from file)\n", jmax);
        status = EGADS_NODATA;
        goto cleanup;
    }

    /* save the array size from the file */
    Imax[numUdp] = imax;
    Jmax[numUdp] = jmax;

    /* allocate necessary arrays */
    Cp[numUdp] = (double*) EG_alloc(3*imax*jmax*sizeof(double));

    if (Cp[numUdp] == NULL) {
        status = EGADS_MALLOC;
        goto cleanup;
    }

    /* read the data */
    for (j = 0; j < jmax; j++) {
        for (i = 0; i < imax; i++) {
            count = fscanf(fp, "%lf %lf %lf", &(Cp[numUdp][3*(i+j*imax)  ]),
                                              &(Cp[numUdp][3*(i+j*imax)+1]),
                                              &(Cp[numUdp][3*(i+j*imax)+2]));
            if (count != 3) {
                status = EGADS_NODATA;
                goto cleanup;
            }
        }
    }

    /* close the file */
    fclose(fp);

    /* create WireBody */
    if (Jmax[numUdp] == 1) {
        nipat = (imax - 1) / 3;

        if (Debug[0]) {
            printf("nipat=%d\n", nipat);
        }

        periodic = 1;
        i0 = 0;
        i1 = imax - 1;
        if (fabs(Cp[numUdp][3*i0  ] - Cp[numUdp][3*i1  ]) > tol ||
            fabs(Cp[numUdp][3*i0+1] - Cp[numUdp][3*i1+1]) > tol ||
            fabs(Cp[numUdp][3*i0+2] - Cp[numUdp][3*i1+2]) > tol   ) {
            periodic = 0;
        }

        /* temporary storage used during construction */
        enodes = (ego*) EG_alloc((nipat+1)*sizeof(ego));
        eedges = (ego*) EG_alloc((nipat  )*sizeof(ego));
        senses = (int*) EG_alloc((nipat  )*sizeof(int));

        if (enodes == NULL || eedges == NULL || senses == NULL) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        /* make the Nodes */
        for (ipat = 0; ipat <= nipat; ipat++) {
            if (ipat == nipat && periodic == 1) {
                enodes[ipat] = enodes[0];
                CHECK_STATUS(EG_copyTopology(NODE));
            } else {
                status = EG_makeTopology(context, NULL, NODE, 0,
                                         &(Cp[numUdp][9*ipat]), 0, NULL, NULL, &(enodes[ipat]));
                CHECK_STATUS(EG_makeTopoloogy(NODE));
            }
        }

        /* loop through the patches of the bezier splines */
        for (ipat = 0; ipat < nipat; ipat++) {

            /* make the (cubic) bezier curve */
            header[0] = 0;
            header[1] = 3;
            header[2] = 4;

            status = EG_makeGeometry(context, CURVE, BEZIER, NULL,
                                     header, &(Cp[numUdp][9*ipat]), &ecurve);
            CHECK_STATUS(EG_makeGeometry(CURVE));

            status = EG_getRange(ecurve, trange, &idum);
            CHECK_STATUS(EG_getRange);

            /* make the Edge */
            status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                     trange, 2, &(enodes[ipat]), NULL, &(eedges[ipat]));
            CHECK_STATUS(EG_makeTopology(EDGE));

            senses[ipat] = SFORWARD;
        }

        /* make a Loop */
        if (periodic == 1) {
            status = EG_makeTopology(context, NULL, LOOP, CLOSED,
                                     NULL, nipat, eedges, senses, &eloop);
            CHECK_STATUS(EG_makeTopology(LOOP-CLOSED));
        } else {
            status = EG_makeTopology(context, NULL, LOOP, OPEN,
                                     NULL, nipat, eedges, senses, &eloop);
            CHECK_STATUS(EG_makeTopology(LOOP-OPEN));
        }

        /* make a WireBody */
        status = EG_makeTopology(context, NULL, BODY, WIREBODY,
                                 NULL, 1, &eloop, NULL, ebody);
        CHECK_STATUS(EG_makeTopology(WIREBODY));

        /* remember this model (body) */
        ebodys[numUdp] = *ebody;

    /* create SolidBody (if closed) or SheetBody (if open) */
    } else {
        nipat = (imax - 1) / 3;
        njpat = (jmax - 1) / 3;

        if (Debug[0]) {
            printf("nipat=%d,  njpat=%d\n", nipat, njpat);
        }

        /* temporary storage used during construction */
        enodes = (ego*) EG_alloc(  (nipat+1)*(njpat+1)*sizeof(ego));
        eedges = (ego*) EG_alloc(2*(nipat+1)*(njpat+1)*sizeof(ego));    /* actually bigger than needed */
        esense = (int*) EG_alloc(2*(nipat+1)*(njpat+1)*sizeof(int));    /* actually bigger than needed */
        efaces = (ego*) EG_alloc(  (nipat  )*(njpat  )*sizeof(ego));
        senses = (int*) EG_alloc(           4         *sizeof(int));

        if (enodes == NULL || eedges == NULL || esense == NULL ||
            efaces == NULL || senses == NULL                     ) {
            status = EGADS_MALLOC;
            goto cleanup;
        }

        /* make (or copy) the Nodes */
        inode = 0;
        for (j0 = 0; j0 <= njpat; j0++) {
            for (i0 = 0; i0 <= nipat; i0++) {
                icp0    = 3 * (i0 + j0 * imax);
                newnode = 1;

                /* look to see if we have a Node at this location already */
                for (j1 = 0; j1 <= njpat; j1++) {
                    for (i1 = 0; i1 <= nipat; i1++) {
                        icp1 = 3 * (i1 + j1 * imax);

                        if (i1 == i0 && j1 == j0) {
                            break;
                        } else if (fabs(Cp[numUdp][3*icp0  ]-Cp[numUdp][3*icp1  ]) < tol &&
                                   fabs(Cp[numUdp][3*icp0+1]-Cp[numUdp][3*icp1+1]) < tol &&
                                   fabs(Cp[numUdp][3*icp0+2]-Cp[numUdp][3*icp1+2]) < tol   ) {
                            enodes[inode] = enodes[i1+j1*(nipat+1)];
                            CHECK_STATUS(EG_copyTopology(NODE));
                            newnode = 0;
                            break;
                        }
                    }
                    if (newnode == 0 || (i1 == i0 && j1 == j0)) {
                        break;
                    }
                }

                if (newnode > 0) {
                    status = EG_makeTopology(context, NULL, NODE, 0,
                                             &(Cp[numUdp][3*icp0]), 0, NULL, NULL, &(enodes[inode]));
                    CHECK_STATUS(EG_makeTopology(NODE));
                }

                if (Debug[0]) {
                    printf("inode=%d, i0=%d, j0=%d, enodes=%lx\n",
                           inode, i0, j0, (long)(enodes[inode]));
                }

                inode++;
            }
        }

        /* make (or copy) the j=constant Edges */
        if (Debug[0]) {
            printf("Making j=constant Edges...\n");
        }

        iedge = 0;
        for (jpat = 0; jpat <= njpat; jpat++) {
            for (ipat = 0; ipat < nipat; ipat++) {
                newedge  = 1;
                etemp[0] = enodes[ipat  +jpat*(nipat+1)];
                etemp[1] = enodes[ipat+1+jpat*(nipat+1)];

                jedge = 0;
                for (j1 = 0; j1 <= njpat; j1++) {
                    for (i1 = 0; i1 < nipat; i1++) {
                        etemp[2] = enodes[i1  +j1*(nipat+1)];
                        etemp[3] = enodes[i1+1+j1*(nipat+1)];

                        if (jedge >= iedge) {
                            break;

                        /* periodic Edge? */
                        } else if (etemp[0] != etemp[1] && etemp[0] == etemp[2] && etemp[1] == etemp[3]) {
                            eedges[iedge] =  eedges[jedge];
                            esense[iedge] = +esense[jedge];
                            CHECK_STATUS(EG_copyGeometry(EDGE));
                            newedge = 0;
                            break;

                        /* anti-periodic edge? */
                        } else if (etemp[0] != etemp[1] && etemp[0] == etemp[3] && etemp[1] == etemp[2]) {
                            eedges[iedge] =  eedges[jedge];
                            esense[iedge] = -esense[jedge];
                            CHECK_STATUS(EG_copyGeometry(EDGE));
                            newedge = 0;
                            break;
                        }
                        jedge++;
                    }
                    if (newedge == 0 || jedge >= iedge) break;
                }

                if (newedge == 0) {
                    /* copied above */

                } else if (etemp[0] == etemp[1]) {
                    trange[0] = 0;
                    trange[1] = 1;

                    status = EG_makeTopology(context, NULL, EDGE, DEGENERATE,
                                             trange, 1, etemp, NULL, &(eedges[iedge]));
                    CHECK_STATUS(EG_makeTopology(EDGE));
                    esense[iedge] = SFORWARD;
                } else {
                    /* make the (cubic) bezier curve */
                    header[0] = 0;
                    header[1] = 3;
                    header[2] = 4;

                    icp = 3 * (ipat + jpat * imax);
                    cp[ 0] = Cp[numUdp][3*icp  ];
                    cp[ 1] = Cp[numUdp][3*icp+1];
                    cp[ 2] = Cp[numUdp][3*icp+2]; icp++;

                    cp[ 3] = Cp[numUdp][3*icp  ];
                    cp[ 4] = Cp[numUdp][3*icp+1];
                    cp[ 5] = Cp[numUdp][3*icp+2]; icp++;

                    cp[ 6] = Cp[numUdp][3*icp  ];
                    cp[ 7] = Cp[numUdp][3*icp+1];
                    cp[ 8] = Cp[numUdp][3*icp+2]; icp++;

                    cp[ 9] = Cp[numUdp][3*icp  ];
                    cp[10] = Cp[numUdp][3*icp+1];
                    cp[11] = Cp[numUdp][3*icp+2];

                    status = EG_makeGeometry(context, CURVE, BEZIER, NULL,
                                             header, cp, &ecurve);
                    CHECK_STATUS(EG_makeGeometry(CURVE));

                    status = EG_getRange(ecurve, trange, &periodic);
                    CHECK_STATUS(EG_getRange);


                    /* make the Edge */
                    status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                             trange, 2, etemp, NULL, &(eedges[iedge]));
                    CHECK_STATUS(EG_makeTopology(EDGE));
                    esense[iedge] = SFORWARD;
                }

                if (Debug[0]) {
                    printf("iedge=%d, ipat=%d, jpat=%d, eedges=%lx, esense=%d\n",
                           iedge, ipat, jpat, (long)(eedges[iedge]), esense[iedge]);
                }

                iedge++;
            }
        }

        /* make (or copy) the i=constant Edges */
        if (Debug[0]) {
            printf("Making i=constant Edges...\n");
        }

        for (jpat = 0; jpat < njpat; jpat++) {
            for (ipat = 0; ipat <= nipat; ipat++) {
                newedge  = 1;
                etemp[0] = enodes[ipat+(jpat  )*(nipat+1)];
                etemp[1] = enodes[ipat+(jpat+1)*(nipat+1)];

                jedge = nipat * (njpat + 1);
                for (j1 = 0; j1 < njpat; j1++) {
                    for (i1 = 0; i1 <= nipat; i1++) {
                        etemp[2] = enodes[i1+(j1  )*(nipat+1)];
                        etemp[3] = enodes[i1+(j1+1)*(nipat+1)];

                        if (jedge >= iedge) {
                            break;

                        /* periodic Edge? */
                        } else if (etemp[0] != etemp[1] && etemp[0] == etemp[2] && etemp[1] == etemp[3]) {
                            eedges[iedge] =  eedges[jedge];
                            esense[iedge] = +esense[jedge];
                            CHECK_STATUS(EG_copyGeometry(EDGE));
                            newedge = 0;
                            break;

                        /* anti-periodic edge? */
                        } else if (etemp[0] != etemp[1] && etemp[0] == etemp[3] && etemp[1] == etemp[2]) {
                            eedges[iedge] =  eedges[jedge];
                            esense[iedge] = -esense[jedge];
                            CHECK_STATUS(EG_copyGeometry(EDGE));
                            newedge = 0;
                            break;
                        }
                        jedge++;
                    }
                    if (newedge == 0 || jedge >= iedge) break;
                }

                if (newedge == 0) {
                    /* copied above */

                } else if (etemp[0] == etemp[1]) {
                    trange[0] = 0;
                    trange[1] = 1;

                    status = EG_makeTopology(context, NULL, EDGE, DEGENERATE,
                                             trange, 1, etemp, NULL, &(eedges[iedge]));
                    CHECK_STATUS(EG_makeTopology(EDGE));
                    esense[iedge] = SFORWARD;
                } else {
                    /* make the (cubic) bezier curve */
                    header[0] = 0;
                    header[1] = 3;
                    header[2] = 4;

                    icp = 3 * (ipat + jpat * imax);
                    cp[ 0] = Cp[numUdp][3*icp  ];
                    cp[ 1] = Cp[numUdp][3*icp+1];
                    cp[ 2] = Cp[numUdp][3*icp+2]; icp+=imax;

                    cp[ 3] = Cp[numUdp][3*icp  ];
                    cp[ 4] = Cp[numUdp][3*icp+1];
                    cp[ 5] = Cp[numUdp][3*icp+2]; icp+=imax;

                    cp[ 6] = Cp[numUdp][3*icp  ];
                    cp[ 7] = Cp[numUdp][3*icp+1];
                    cp[ 8] = Cp[numUdp][3*icp+2]; icp+=imax;

                    cp[ 9] = Cp[numUdp][3*icp  ];
                    cp[10] = Cp[numUdp][3*icp+1];
                    cp[11] = Cp[numUdp][3*icp+2];

                    status = EG_makeGeometry(context, CURVE, BEZIER, NULL,
                                             header, cp, &ecurve);
                    CHECK_STATUS(EG_makeGeometry(CURVE));

                    status = EG_getRange(ecurve, trange, &periodic);
                    CHECK_STATUS(EG_getRange);

                    /* make the Edge */
                    status = EG_makeTopology(context, ecurve, EDGE, TWONODE,
                                             trange, 2, etemp, NULL, &(eedges[iedge]));
                    CHECK_STATUS(EG_makeTopology(EDGE));
                    esense[iedge] = SFORWARD;
                }

                if (Debug[0]) {
                    printf("iedge=%d, ipat=%d, jpat=%d, eedges=%lx, esense=%d\n",
                           iedge, ipat, jpat, (long)(eedges[iedge]), esense[iedge]);
                }

                iedge++;
            }
        }

        /* make each of the (bicubic) bezier surface patches */
        if (Debug[0]) {
            printf("Making Faces...\n");
        }

        iface = 0;
        for (jpat = 0; jpat < njpat; jpat++) {
            for (ipat = 0; ipat < nipat; ipat++) {

                /* make the (cubic) bezier curve */
                header[0] = 0;
                header[1] = 3;
                header[2] = 4;
                header[3] = 3;
                header[4] = 4;

                icp = 3 * (ipat + jpat * imax);
                cp[ 0] = Cp[numUdp][3*icp  ];
                cp[ 1] = Cp[numUdp][3*icp+1];
                cp[ 2] = Cp[numUdp][3*icp+2]; icp++;

                cp[ 3] = Cp[numUdp][3*icp  ];
                cp[ 4] = Cp[numUdp][3*icp+1];
                cp[ 5] = Cp[numUdp][3*icp+2]; icp++;

                cp[ 6] = Cp[numUdp][3*icp  ];
                cp[ 7] = Cp[numUdp][3*icp+1];
                cp[ 8] = Cp[numUdp][3*icp+2]; icp++;

                cp[ 9] = Cp[numUdp][3*icp  ];
                cp[10] = Cp[numUdp][3*icp+1];
                cp[11] = Cp[numUdp][3*icp+2]; icp+=imax-3;

                cp[12] = Cp[numUdp][3*icp  ];
                cp[13] = Cp[numUdp][3*icp+1];
                cp[14] = Cp[numUdp][3*icp+2]; icp++;

                cp[15] = Cp[numUdp][3*icp  ];
                cp[16] = Cp[numUdp][3*icp+1];
                cp[17] = Cp[numUdp][3*icp+2]; icp++;

                cp[18] = Cp[numUdp][3*icp  ];
                cp[19] = Cp[numUdp][3*icp+1];
                cp[20] = Cp[numUdp][3*icp+2]; icp++;

                cp[21] = Cp[numUdp][3*icp  ];
                cp[22] = Cp[numUdp][3*icp+1];
                cp[23] = Cp[numUdp][3*icp+2]; icp+=imax-3;

                cp[24] = Cp[numUdp][3*icp  ];
                cp[25] = Cp[numUdp][3*icp+1];
                cp[26] = Cp[numUdp][3*icp+2]; icp++;

                cp[27] = Cp[numUdp][3*icp  ];
                cp[28] = Cp[numUdp][3*icp+1];
                cp[29] = Cp[numUdp][3*icp+2]; icp++;

                cp[30] = Cp[numUdp][3*icp  ];
                cp[31] = Cp[numUdp][3*icp+1];
                cp[32] = Cp[numUdp][3*icp+2]; icp++;

                cp[33] = Cp[numUdp][3*icp  ];
                cp[34] = Cp[numUdp][3*icp+1];
                cp[35] = Cp[numUdp][3*icp+2]; icp+=imax-3;

                cp[36] = Cp[numUdp][3*icp  ];
                cp[37] = Cp[numUdp][3*icp+1];
                cp[38] = Cp[numUdp][3*icp+2]; icp++;

                cp[39] = Cp[numUdp][3*icp  ];
                cp[40] = Cp[numUdp][3*icp+1];
                cp[41] = Cp[numUdp][3*icp+2]; icp++;

                cp[42] = Cp[numUdp][3*icp  ];
                cp[43] = Cp[numUdp][3*icp+1];
                cp[44] = Cp[numUdp][3*icp+2]; icp++;

                cp[45] = Cp[numUdp][3*icp  ];
                cp[46] = Cp[numUdp][3*icp+1];
                cp[47] = Cp[numUdp][3*icp+2];

                status = EG_makeGeometry(context, SURFACE, BEZIER, NULL,
                                         header, cp, &esurf);
                CHECK_STATUS(EG_makeGeometry(SURFACE));

                /* make the Loop */
                etemp[ 0] =  eedges[ipat  + jpat * nipat];
                senses[0] =  esense[ipat  + jpat * nipat];

                etemp[ 1] =  eedges[ipat+1 + jpat * (nipat+1) + nipat * (njpat+1)];
                senses[1] =  esense[ipat+1 + jpat * (nipat+1) + nipat * (njpat+1)];

                etemp[ 2] =  eedges[ipat + (jpat+1) * nipat];
                senses[2] = -esense[ipat + (jpat+1) * nipat];

                etemp[ 3] =  eedges[ipat + jpat * (nipat+1) + nipat * (njpat+1)];
                senses[3] = -esense[ipat + jpat * (nipat+1) + nipat * (njpat+1)];

                if (senses[0] > 0) {
                    data[0] = 0;  data[1] = 0;  data[2] = +1;  data[3] = 0;
                } else {
                    data[0] = 1;  data[1] = 0;  data[2] = -1;  data[3] = 0;
                }
                status = EG_makeGeometry(context, PCURVE, LINE, esurf,
                                         NULL, data, &(etemp[4]));
                CHECK_STATUS(EG_makeGeometry(PCURVE));
                 
                if (senses[1] > 0) {
                    data[0] = 1;  data[1] = 0;  data[2] = 0;  data[3] = +1;
                } else {
                    data[0] = 1;  data[1] = 1;  data[2] = 0;  data[3] = -1;
                }
                status = EG_makeGeometry(context, PCURVE, LINE, esurf,
                                         NULL, data, &(etemp[5]));
                CHECK_STATUS(EG_makeGeometry(PCURVE));
                 
                if (senses[2] < 0) {
                    data[0] = 0;  data[1] = 1;  data[2] = +1;  data[3] = 0;
                } else {
                    data[0] = 1;  data[1] = 1;  data[2] = -1;  data[3] = 0;
                }
                status = EG_makeGeometry(context, PCURVE, LINE, esurf,
                                         NULL, data, &(etemp[6]));
                CHECK_STATUS(EG_makeGeometry(PCURVE));
                 
                if (senses[3] < 0) {
                    data[0] = 0;  data[1] = 0;  data[2] = 0;  data[3] = +1;
                } else {
                    data[0] = 0;  data[1] = 1;  data[2] = 0;  data[3] = -1;
                }
                status = EG_makeGeometry(context, PCURVE, LINE, esurf,
                                         NULL, data, &(etemp[7]));
                CHECK_STATUS(EG_makeGeometry(PCURVE));

                status = EG_makeTopology(context, esurf, LOOP, CLOSED,
                                         NULL, 4, etemp, senses, &eloop);
                CHECK_STATUS(EG_makeTopology(LOOP));

                /* make the Face */
                senses[0] = SFORWARD;
                status = EG_makeTopology(context, esurf, FACE, SFORWARD,
                                         NULL, 1, &eloop, senses, &(efaces[iface]));
                CHECK_STATUS(EG_makeTopology);

                if (Debug[0]) {
                    printf("iface=%d, ipat=%d, jpat=%d, efaces=%lx\n", iface, ipat, jpat, (long)(efaces[iface]));
                }

                iface++;
            }
        }

        /* make a Shell (assume CLOSED, but fixed below) */
        status = EG_makeTopology(context, NULL, SHELL, CLOSED,
                                 NULL, iface, efaces, NULL, &eshell);
        CHECK_STATUS(EG_makeTopology(SHELL));

        /* determine if Shell is open or closed */
        status = EG_getTopology(eshell, &eref, &oclass, &mtype,
                                trange, &nchild, &echild, &senses2);
        CHECK_STATUS(EG_getTopology);

        if (mtype == CLOSED) {
            /* make a SolidBody */
            status = EG_makeTopology(context, NULL, BODY, SOLIDBODY,
                                     NULL, 1, &eshell, NULL, ebody);
            CHECK_STATUS(EG_makeTopology(SOLIDBODY));
        } else {
            /* make a SheetBody */
            status = EG_makeTopology(context, NULL, BODY, SHEETBODY,
                                     NULL, 1, &eshell, NULL, ebody);
            CHECK_STATUS(EG_makeTopology(SHEETBODY));
        }
    }


cleanup:
    if (enodes != NULL) EG_free(enodes);
    if (eedges != NULL) EG_free(eedges);
    if (esense != NULL) EG_free(esense);
    if (efaces != NULL) EG_free(efaces);
    if (senses != NULL) EG_free(senses);

    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpMesh - return mesh associated with primitive                    *
 *                                                                      *
 ************************************************************************
 */

int
udpMesh(ego    ebody,
        int    iMesh,
        int    *imax,
        int    *jmax,
        int    *kmax,
        double **mesh)
{
    int    iudp, judp;

    *imax = 0;
    *jmax = 0;
    *kmax = 0;
    *mesh = NULL;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (iMesh != 0) {
        printf(" udpMesh: iMesh %d != 0\n", iMesh);
    }

    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,
               char   *vname,
  /*@unused@*/ int    npts,
  /*@unused@*/ int    *fIndices,
  /*@unused@*/ double *uvs,
  /*@unused@*/ double *dxname)
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpStepSize - return sensitivity step size of finite differencing  *
 *                                                                      *
 ************************************************************************
 */

int
udpStepSize(ego    ebody,
            char   *vname,
            double *delta)
{
    int iudp, judp;

    *delta = 0;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}

#endif
