/*
 ************************************************************************
 *                                                                      *
 * udpBox -- udp file to generate a box                                 *
 *                                                                      *
 *            Written by Bridget Dixon @ Syracuse University            *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                      John Dannenhoffer @ Syracuse University         *
 *                                                                      *
 ************************************************************************
 */

#ifdef GEOM_EGADS

/*
 * Copyright (C) 2013  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef WIN32
    #define snprintf _snprintf
#endif

#include "egads.h"

/* ADDITIONAL ATTRIBUTE TYPE */
#define ATTRREALSEN 4

#define TWOPI 6.2831853071795862319959269

/* static variables used to hold info between function calls */

/* number of Ellipse UDPs */
static int    numUdp  = 0;

/* arrays to hold info for each of the UDPs */
static ego    *ebodys   = NULL;
static double *Dx       = NULL;
static double *Dy       = NULL;
static double *Dz       = NULL;
static double *Rad      = NULL;


/*
 ************************************************************************
 *                                                                      *
 *   udpErrorStr - print error string                                   *
 *                                                                      *
 ************************************************************************
 */

static /*@null@*/ char *
udpErrorStr(int stat)
{
    char *string;

    string = EG_alloc(25*sizeof(char));
    if (string == NULL) {
        return string;
    }
    snprintf(string, 25, "EGADS status = %d", stat);

    return string;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpInitialize - initialize and get info about the list of arguments*
 *                                                                      *
 ************************************************************************
 */

int
udpInitialize(int    *nArgs,
              char   ***namex,
              int    **typex,
              int    **idefaulx,
              double **ddefaulx)
{
    int    *type, *idefault;
    char   **name;
    double *ddefault;

    /* make the arrays */
    ebodys = (ego    *) EG_alloc(sizeof(ego    ));
    Dx     = (double *) EG_alloc(sizeof(double ));
    Dy     = (double *) EG_alloc(sizeof(double ));
    Dz     = (double *) EG_alloc(sizeof(double ));
    Rad    = (double *) EG_alloc(sizeof(double ));

    if (ebodys == NULL || Dx == NULL || Dy == NULL || Dz == NULL || Rad == NULL) {
        return EGADS_MALLOC;
    }

    /* initialize the array elements that will hold the "current" settings */
    ebodys[0] = NULL;
    Dx[    0] = 0;
    Dy[    0] = 0;
    Dz[    0] = 0;
    Rad[   0] = 0;

    /* set up returns that describe the UDP */
    *nArgs    = 4;
    *namex    = NULL;
    *typex    = NULL;
    *idefaulx = NULL;
    *ddefaulx = NULL;

    *namex = name = (char **) EG_alloc((*nArgs)*sizeof(char *));
    if (*namex == NULL) {
        return EGADS_MALLOC;
    }

    *typex = type = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*typex == NULL) {
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *idefaulx = idefault = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*idefaulx == NULL) {
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *ddefaulx = ddefault = (double *) EG_alloc((*nArgs)*sizeof(double));
    if (*ddefaulx == NULL) {
        EG_free(*idefaulx);
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    name[0]     = "dx";
    type[0]     = ATTRREAL;
    idefault[0] = 0;
    ddefault[0] = 0;

    name[1]     = "dy";
    type[1]     = ATTRREAL;
    idefault[1] = 0;
    ddefault[1] = 0;

    name[2]     = "dz";
    type[2]     = ATTRREAL;
    idefault[2] = 0;
    ddefault[2] = 0;

    name[3]     = "rad";
    type[3]     = ATTRREAL;
    idefault[3] = 0;
    ddefault[3] = 0;

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpReset - reset the arguments to their defaults                   *
 *                                                                      *
 ************************************************************************
 */

int
udpReset(int flag)
{

    /* reset the "current" settings */
    if (flag == 0) {
        Dx[ 0] = 0;
        Dy[ 0] = 0;
        Dz[ 0] = 0;
        Rad[0] = 0;

    /* called when closing up */
    } else {
        EG_free(ebodys); ebodys = NULL;
        EG_free(Dx    ); Dx     = NULL;
        EG_free(Dy    ); Dy     = NULL;
        EG_free(Dz    ); Dz     = NULL;
        EG_free(Rad   ); Rad    = NULL;

        numUdp = 0;
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSet - set an argument                                           *
 *                                                                      *
 ************************************************************************
 */

int
udpSet(char *name,
       char *value)
{
    int  i, len;
    char *pEnd, lowername[257];

    if        (name  == NULL) {
        return EGADS_NONAME;
    } else if (strlen(name) == 0 || strlen(name) > 255) {
        return EGADS_NONAME;
    } else if (value == NULL) {
        return EGADS_NULLOBJ;
    }

    for (i = 0; i <= strlen(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    len = strlen(value);
    if (len == 0) {
        return EGADS_NODATA;
    }

    if        (strcmp(lowername, "dx") == 0) {
        Dx[0] = strtod(value, &pEnd);
        if (Dx[0] < 0) {
            printf(" udpSet: dx = %f < 0\n", Dx[0]);
            return EGADS_RANGERR;
        }

    } else if (strcmp(lowername, "dy") == 0) {
        Dy[0] = strtod(value, &pEnd);
        if (Dy[0] < 0) {
            printf(" udpSet: dy = %f < 0\n", Dy[0]);
            return EGADS_RANGERR;
        }

    } else if (strcmp(lowername, "dz") == 0) {
        Dz[0] = strtod(value, &pEnd);
        if (Dz[0] < 0) {
            printf(" udpSet: dz = %f < 0\n", Dz[0]);
            return EGADS_RANGERR;
        }

    } else if (strcmp(lowername, "rad") == 0) {
        Rad[0] = strtod(value, &pEnd);
        if (Rad[0] < 0) {
            printf(" udpSet: rad = %f < 0\n", Rad[0]);
            return EGADS_RANGERR;
        }

    } else {
        printf(" udpSet: Parameter \"%s\" not known (should be dx, dy, dz, or rad)\n", name);
        return EGADS_INDEXERR;

    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,
           ego  *ebody,
           int  *nMesh,
           char **string)
{
    int     status = EGADS_SUCCESS;
    int     wire, sense[8], nedges;
    double  node1[3], node2[3], node3[3], node4[3], node5[3], node6[3], node7[3], node8[3];
    double  cent1[3], cent2[3], cent3[3], cent4[3], Xaxis[3], Yaxis[3];
    double  data[18], trange[2];
    ego     enodes[9], ecurve, eedges[8], eloop, eface, etemp, *eedges2;

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* increment number of UDPs */
    numUdp++;

    /* increase the arrays to make room for the new UDP */
    ebodys = (ego    *) EG_reall(ebodys, (numUdp+1)*sizeof(ego    ));
    Dx     = (double *) EG_reall(Dx,     (numUdp+1)*sizeof(double ));
    Dy     = (double *) EG_reall(Dy,     (numUdp+1)*sizeof(double ));
    Dz     = (double *) EG_reall(Dz,     (numUdp+1)*sizeof(double ));
    Rad    = (double *) EG_reall(Rad,    (numUdp+1)*sizeof(double ));

    if (ebodys == NULL || Dx == NULL || Dy == NULL || Dz == NULL || Rad == NULL) {
        return EGADS_MALLOC;
    }

    Dx[numUdp]  = Dx[0];
    Dy[numUdp]  = Dy[0];
    Dz[numUdp]  = Dz[0];
    Rad[numUdp] = Rad[0];

    /* make sure that at least one of the dimensions is positive */
    if (Dx[0] <= 0 && Dy[0] <= 0 && Dz[0] <= 0) {
        status = EGADS_GEOMERR;
        *string = udpErrorStr(status);
        goto cleanup;

    /* check for 3D solid body (and make if requested) */
    } else if (Dx[0] > 0 && Dy[0] > 0 && Dz[0] > 0)  {

        /* make sure that radius is not too big */
        if (2*Rad[0] >= Dx[0] || 2*Rad[0] >= Dy[0] || 2*Rad[0] >= Dz[0]) {
            printf(" udpExecute: radius cannot be greater than half of any side length\n");
            status = EGADS_GEOMERR;
            *string = udpErrorStr(status);
            goto cleanup;
        }

        data[0] = -Dx[0] / 2;
        data[1] = -Dy[0] / 2;
        data[2] = -Dz[0] / 2;
        data[3] =  Dx[0];
        data[4] =  Dy[0];
        data[5] =  Dz[0];

        /* Make Solid Body */
        status = EG_makeSolidBody(context, BOX, data, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* apply rounded edges (if required) */
        if (Rad[0] > 0) {
            etemp = *ebody;

            status = EG_getBodyTopos(etemp, NULL, EDGE, &nedges, &eedges2);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            status = EG_filletBody(etemp, nedges, eedges2, Rad[0], ebody);
            if (status != EGADS_SUCCESS) {
                *string = udpErrorStr(status);
                goto cleanup;
            }

            EG_free(eedges2);
        }

        /* remember this model (body) */
        ebodys[numUdp] = *ebody;
        goto cleanup;
    }

    /* otherwise this is either a 1D wirebody or a 2D facebody */
    wire = 0;            /* flag to tell if wire or face */

    /* geometry start */
    node1[0] = -Dx[0] / 2;
    node1[1] = -Dy[0] / 2;
    node1[2] = -Dz[0] / 2;

    /* wire body dz > 0 */
    if        (Dx[0] == 0 && Dy[0] == 0) {
        wire = 1;

        node2[0] =    0;
        node2[1] =    0;
        node2[2] = Dz[0] / 2;

    /* wire body dy > 0 */
    } else if (Dx[0] == 0 && Dz[0] == 0) {
        wire = 1;

        node2[0] =    0;
        node2[1] = Dy[0] / 2;
        node2[2] =    0;

    /* wire body dx > 0 */
    } else if (Dy[0] == 0 && Dz[0] == 0) {
        wire = 1;

        node2[0] = Dx[0] / 2;
        node2[1] =    0;
        node2[2] =    0;

    /* face in the x-y plane */
    } else if (Dz[0] == 0) {
        node2[0] = +Dx[0] / 2;
        node2[1] = -Dy[0] / 2;
        node2[2] =     0;

        node3[0] = +Dx[0] / 2;
        node3[1] = +Dy[0] / 2;
        node3[2] =     0;

        node4[0] = -Dx[0] / 2;
        node4[1] = +Dy[0] / 2;
        node4[2] =     0;

    /* face in the y-z plane */
    } else if (Dx[0] == 0) {
        node2[0] =     0;
        node2[1] = +Dy[0] / 2;
        node2[2] = -Dz[0] / 2;

        node3[0] =     0;
        node3[1] = +Dy[0] / 2;
        node3[2] = +Dz[0] / 2;

        node4[0] =     0;
        node4[1] = -Dy[0] / 2;
        node4[2] = +Dz[0] / 2;

    /* face in the x-z plane */
    } else {
        node2[0] = +Dx[0] / 2;
        node2[1] =     0;
        node2[2] = -Dz[0] / 2;

        node3[0] = +Dx[0] / 2;
        node3[1] =     0;
        node3[2] = +Dz[0] / 2;

        node4[0] = -Dx[0] / 2;
        node4[1] =     0;
        node4[2] = +Dz[0] / 2;
    }

    /* make 1D wire body */
    if (wire == 1) {

        /* check to ensure that rad was not set */
        if (Rad[0] > 0) {
            printf(" udpExecute: rad cannot be set for wirebody\n");
            status = EGADS_GEOMERR;
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Nodes */
        status = EG_makeTopology(context, NULL, NODE, 0, node1, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node2, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 1 */
        data[0] = node1[0];
        data[1] = node1[1];
        data[2] = node1[2];
        data[3] = node2[0] - node1[0];
        data[4] = node2[1] - node1[1];
        data[5] = node2[2] - node1[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node1, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node2, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Loop from this Edge */
        sense[0] = SFORWARD;
        sense[1] = SFORWARD;
        status = EG_makeTopology(context, NULL, LOOP, OPEN, NULL, 1, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create the WireBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, WIREBODY, NULL, 1, &eloop, NULL, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* remember this model (body) */
        ebodys[numUdp] = *ebody;
        goto cleanup;

    /* make 2D face body (without rounded corners) */
    } else if (Rad[0] == 0) {

        /* make Nodes */
        status = EG_makeTopology(context, NULL, NODE, 0, node1, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node2, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node3, 0, NULL, NULL, &(enodes[2]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node4, 0, NULL, NULL, &(enodes[3]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* first and last Nodes are the same */
        enodes[4] = enodes[0];

        /* make the Line 1 */
        data[0] = node1[0];
        data[1] = node1[1];
        data[2] = node1[2];
        data[3] = node2[0] - node1[0];
        data[4] = node2[1] - node1[1];
        data[5] = node2[2] - node1[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node1, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node2, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 2 */
        data[0] = node2[0];
        data[1] = node2[1];
        data[2] = node2[2];
        data[3] = node3[0] - node2[0];
        data[4] = node3[1] - node2[1];
        data[5] = node3[2] - node2[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node2, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node3, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[1]), NULL, &(eedges[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 3 */
        data[0] = node3[0];
        data[1] = node3[1];
        data[2] = node3[2];
        data[3] = node4[0] - node3[0];
        data[4] = node4[1] - node3[1];
        data[5] = node4[2] - node3[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node3, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node4, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[2]), NULL, &(eedges[2]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 4 */
        data[0] = node4[0];
        data[1] = node4[1];
        data[2] = node4[2];
        data[3] = node1[0] - node4[0];
        data[4] = node1[1] - node4[1];
        data[5] = node1[2] - node4[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node4, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node1, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[3]), NULL, &(eedges[3]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Loop from this Edge */
        sense[0] = SFORWARD;
        sense[1] = SFORWARD;
        sense[2] = SFORWARD;
        sense[3] = SFORWARD;

        status = EG_makeTopology(context, NULL, LOOP, CLOSED, NULL, 4, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Face from the loop */
        status = EG_makeFace(eloop, SFORWARD, NULL, &eface);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create the FaceBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, FACEBODY, NULL, 1, &eface, NULL, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* remember this model (body) */
        ebodys[numUdp] = *ebody;
        goto cleanup;

    /* make 2D face body (with rounded corners) */
    } else {

        /* make Nodes */
        if (Dx[0] > 0 && Dy[0] > 0 && Dz[0] == 0){

            /* make sure that radius is not too big */
            if (2*Rad[0] >= Dx[0] || 2*Rad[0] >= Dy[0]) {
                printf(" udpExecute: radius cannot be greater than half of any side length\n");
                status = EGADS_GEOMERR;
                *string = udpErrorStr(status);
                goto cleanup;
            }

            /* in x-y plane */
            node1[0] = -Dx[0] / 2 + Rad[0];
            node1[1] = -Dy[0] / 2;
            node1[2] =  0;

            node2[0] = +Dx[0] / 2 - Rad[0];
            node2[1] = -Dy[0] / 2;
            node2[2] =  0;

            node3[0] = +Dx[0] / 2;
            node3[1] = -Dy[0] / 2 + Rad[0];
            node3[2] =  0;

            node4[0] = +Dx[0] / 2;
            node4[1] = +Dy[0] / 2 - Rad[0];
            node4[2] =  0;

            node5[0] = +Dx[0] / 2 - Rad[0];
            node5[1] = +Dy[0] / 2;
            node5[2] =  0;

            node6[0] = -Dx[0] / 2 + Rad[0];
            node6[1] = +Dy[0] / 2;
            node6[2] =  0;

            node7[0] = -Dx[0] / 2;
            node7[1] = +Dy[0] / 2 - Rad[0];
            node7[2] =  0;

            node8[0] = -Dx[0] / 2;
            node8[1] = -Dy[0] / 2 + Rad[0];
            node8[2] =  0;

            cent1[0] = +Dx[0] / 2 - Rad[0];
            cent1[1] = -Dy[0] / 2 + Rad[0];
            cent1[2] =  0;

            cent2[0] = +Dx[0] / 2 - Rad[0];
            cent2[1] = +Dy[0] / 2 - Rad[0];
            cent2[2] =  0;

            cent3[0] = -Dx[0] / 2 + Rad[0];
            cent3[1] = +Dy[0] / 2 - Rad[0];
            cent3[2] =  0;

            cent4[0] = -Dx[0] / 2 + Rad[0];
            cent4[1] = -Dy[0] / 2 + Rad[0];
            cent4[2] =  0;

            Xaxis[0] = 1;
            Xaxis[1] = 0;
            Xaxis[2] = 0;

            Yaxis[0] = 0;
            Yaxis[1] = 1;
            Yaxis[2] = 0;

        } else if (Dx[0] > 0 && Dy[0] == 0 && Dz[0] > 0) {

            /* make sure that radius is not too big */
            if (2*Rad[0] >= Dx[0] || 2*Rad[0] >= Dz[0]) {
                printf(" udpExecute: radius cannot be greater than half of any side length\n");
                status = EGADS_GEOMERR;
                *string = udpErrorStr(status);
                goto cleanup;
            }

            /* in x-z plane */
            node1[0] = -Dx[0] / 2 + Rad[0];
            node1[1] =  0;
            node1[2] = -Dz[0] / 2;

            node2[0] = +Dx[0] / 2 - Rad[0];
            node2[1] =  0;
            node2[2] = -Dz[0] / 2;

            node3[0] = +Dx[0] / 2;
            node3[1] =  0;
            node3[2] = -Dz[0] / 2 + Rad[0];

            node4[0] = +Dx[0] / 2;
            node4[1] =  0;
            node4[2] = +Dz[0] / 2 - Rad[0];

            node5[0] = +Dx[0] / 2 - Rad[0];
            node5[1] =  0;
            node5[2] = +Dz[0] / 2;

            node6[0] = -Dx[0] / 2 + Rad[0];
            node6[1] =  0;
            node6[2] = +Dz[0] / 2;

            node7[0] = -Dx[0] / 2;
            node7[1] =  0;
            node7[2] = +Dz[0] / 2 - Rad[0];

            node8[0] = -Dx[0] / 2;
            node8[1] =  0;
            node8[2] = -Dz[0] / 2 + Rad[0];

            cent1[0] = +Dx[0] / 2 - Rad[0];
            cent1[1] =  0;
            cent1[2] = -Dz[0] / 2 + Rad[0];

            cent2[0] = +Dx[0] / 2 - Rad[0];
            cent2[1] =  0;
            cent2[2] = +Dz[0] / 2 - Rad[0];

            cent3[0] = -Dx[0] / 2 + Rad[0];
            cent3[1] =  0;
            cent3[2] = +Dz[0] / 2 - Rad[0];

            cent4[0] = -Dx[0] / 2 + Rad[0];
            cent4[1] =  0;
            cent4[2] = -Dz[0] / 2 + Rad[0];

            Xaxis[0] = 1;
            Xaxis[1] = 0;
            Xaxis[2] = 0;

            Yaxis[0] = 0;
            Yaxis[1] = 0;
            Yaxis[2] = 1;

        } else {

            /* make sure that radius is not too big */
            if (2*Rad[0] >= Dy[0] || 2*Rad[0] >= Dz[0]) {
                printf(" udpExecute: radius cannot be greater than half of any side length\n");
                status = EGADS_GEOMERR;
                *string = udpErrorStr(status);
                goto cleanup;
            }

            /* in y-z plane */
            node1[0] =  0;
            node1[1] = -Dy[0] / 2;
            node1[2] = +Dz[0] / 2 - Rad[0];

            node2[0] =  0;
            node2[1] = -Dy[0] / 2;
            node2[2] = -Dz[0] / 2 + Rad[0];

            node3[0] =  0;
            node3[1] = -Dy[0] / 2 + Rad[0];
            node3[2] = -Dz[0] / 2;

            node4[0] =  0;
            node4[1] = +Dy[0] / 2 - Rad[0];
            node4[2] = -Dz[0] / 2;

            node5[0] =  0;
            node5[1] = +Dy[0] / 2;
            node5[2] = -Dz[0] / 2 + Rad[0];

            node6[0] =  0;
            node6[1] = +Dy[0] / 2;
            node6[2] = +Dz[0] / 2 - Rad[0];

            node7[0] =  0;
            node7[1] = +Dy[0] / 2 - Rad[0];
            node7[2] = +Dz[0] / 2;

            node8[0] =  0;
            node8[1] = -Dy[0] / 2 + Rad[0];
            node8[2] = +Dz[0] / 2;

            cent1[0] =  0;
            cent1[1] = -Dy[0] / 2 + Rad[0];
            cent1[2] = -Dz[0] / 2 + Rad[0];

            cent2[0] =  0;
            cent2[1] = +Dy[0] / 2 - Rad[0];
            cent2[2] = -Dz[0] / 2 + Rad[0];

            cent3[0] =  0;
            cent3[1] = +Dy[0] / 2 - Rad[0];
            cent3[2] = +Dz[0] / 2 - Rad[0];

            cent4[0] =  0;
            cent4[1] = -Dy[0] / 2 + Rad[0];
            cent4[2] = +Dz[0] / 2 - Rad[0];

            Xaxis[0] = 0;
            Xaxis[1] = 1;
            Xaxis[2] = 0;

            Yaxis[0] = 0;
            Yaxis[1] = 0;
            Yaxis[2] = 1;
        }

        /* make the nodes */
        status = EG_makeTopology(context, NULL, NODE, 0, node1, 0, NULL, NULL, &(enodes[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node2, 0, NULL, NULL, &(enodes[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node3, 0, NULL, NULL, &(enodes[2]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node4, 0, NULL, NULL, &(enodes[3]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node5, 0, NULL, NULL, &(enodes[4]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node6, 0, NULL, NULL, &(enodes[5]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node7, 0, NULL, NULL, &(enodes[6]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_makeTopology(context, NULL, NODE, 0, node8, 0, NULL, NULL, &(enodes[7]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* first and last Nodes are the same */
        enodes[8] = enodes[0];

        /* make the Line 1 */
        data[0] = node1[0];
        data[1] = node1[1];
        data[2] = node1[2];
        data[3] = node2[0] - node1[0];
        data[4] = node2[1] - node1[1];
        data[5] = node2[2] - node1[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node1, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node2, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[0]), NULL, &(eedges[0]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make rounded corner 1 */
        data[0] = cent1[0];
        data[1] = cent1[1];
        data[2] = cent1[2];
        data[3] = Xaxis[0];
        data[4] = Xaxis[1];
        data[5] = Xaxis[2];
        data[6] = Yaxis[0];
        data[7] = Yaxis[1];
        data[8] = Yaxis[2];
        data[9] = Rad[0];

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node2, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node3, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        if (trange[0] > trange[1]){
            trange[1] += TWOPI;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[1]), NULL, &(eedges[1]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 2 */
        data[0] = node3[0];
        data[1] = node3[1];
        data[2] = node3[2];
        data[3] = node4[0] - node3[0];
        data[4] = node4[1] - node3[1];
        data[5] = node4[2] - node3[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node3, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node4, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[2]), NULL, &(eedges[2]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make rounded corner 2 */
        data[0] = cent2[0];
        data[1] = cent2[1];
        data[2] = cent2[2];
        data[3] = Xaxis[0];
        data[4] = Xaxis[1];
        data[5] = Xaxis[2];
        data[6] = Yaxis[0];
        data[7] = Yaxis[1];
        data[8] = Yaxis[2];
        data[9] = Rad[0];

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node4, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node5, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        if (trange[0] > trange[1]){
            trange[1] += TWOPI;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[3]), NULL, &(eedges[3]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 3 */
        data[0] = node5[0];
        data[1] = node5[1];
        data[2] = node5[2];
        data[3] = node6[0] - node5[0];
        data[4] = node6[1] - node5[1];
        data[5] = node6[2] - node5[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node5, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node6, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[4]), NULL, &(eedges[4]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make rounded corner 3 */
        data[0] = cent3[0];
        data[1] = cent3[1];
        data[2] = cent3[2];
        data[3] = Xaxis[0];
        data[4] = Xaxis[1];
        data[5] = Xaxis[2];
        data[6] = Yaxis[0];
        data[7] = Yaxis[1];
        data[8] = Yaxis[2];
        data[9] = Rad[0];

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node6, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node7, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        if (trange[0] > trange[1]){
            trange[1] += TWOPI;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[5]), NULL, &(eedges[5]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make the Line 4 */
        data[0] = node7[0];
        data[1] = node7[1];
        data[2] = node7[2];
        data[3] = node8[0] - node7[0];
        data[4] = node8[1] - node7[1];
        data[5] = node8[2] - node7[2];
        status = EG_makeGeometry(context, CURVE, LINE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node7, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node8, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[6]), NULL, &(eedges[6]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make rounded corner 4 */
        data[0] = cent4[0];
        data[1] = cent4[1];
        data[2] = cent4[2];
        data[3] = Xaxis[0];
        data[4] = Xaxis[1];
        data[5] = Xaxis[2];
        data[6] = Yaxis[0];
        data[7] = Yaxis[1];
        data[8] = Yaxis[2];
        data[9] = Rad[0];

        status = EG_makeGeometry(context, CURVE, CIRCLE, NULL, NULL, data, &ecurve);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* get the parameter range */
        status = EG_invEvaluate(ecurve, node8, &(trange[0]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        status = EG_invEvaluate(ecurve, node1, &(trange[1]), data);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        if (trange[0] > trange[1]){
            trange[1] += TWOPI;
        }

        /* make edge */
        status = EG_makeTopology(context, ecurve, EDGE, TWONODE, trange, 2, &(enodes[7]), NULL, &(eedges[7]));
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Loop from this Edge */
        sense[0] = SFORWARD;
        sense[1] = SFORWARD;
        sense[2] = SFORWARD;
        sense[3] = SFORWARD;
        sense[4] = SFORWARD;
        sense[5] = SFORWARD;
        sense[6] = SFORWARD;
        sense[7] = SFORWARD;

        status = EG_makeTopology(context, NULL, LOOP, CLOSED, NULL, 8, eedges, sense, &eloop);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* make Face from the loop */
        status = EG_makeFace(eloop, SFORWARD, NULL, &eface);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* create the FaceBody (which will be returned) */
        status = EG_makeTopology(context, NULL, BODY, FACEBODY, NULL, 1, &eface, NULL, ebody);
        if (status != EGADS_SUCCESS) {
            *string = udpErrorStr(status);
            goto cleanup;
        }

        /* remember this model (body) */
        ebodys[numUdp] = *ebody;
        goto cleanup;
    }

cleanup:
    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpMesh - return mesh associated with primitive                    *
 *                                                                      *
 ************************************************************************
 */

int
udpMesh(ego    ebody,
        int    iMesh,
        int    *imax,
        int    *jmax,
        int    *kmax,
        double **mesh)
{
    int    iudp, judp;

    *imax = 0;
    *jmax = 0;
    *kmax = 0;
    *mesh = NULL;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (iMesh != 0) {
        printf(" udpMesh: iMesh %d != 0\n", iMesh);
    }

    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,
               char   *vname,
  /*@unused@*/ int    npts,
  /*@unused@*/ int    *fIndices,
  /*@unused@*/ double *uvs,
  /*@unused@*/ double *dxdname)
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpStepSize - return sensitivity step size of finite differencing  *
 *                                                                      *
 ************************************************************************
 */

int
udpStepSize(ego    ebody,
            char   *vname,
            double *delta)
{
    int iudp, judp;

    *delta = 0;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}

#endif
