/*
 ************************************************************************
 *                                                                      *
 * udpExtract -- udp file to generate a extract brick                 *
 *                                                                      *
 *            Written by John Dannenhoffer @ Syracuse University        *
 *            Patterned after code written by Bob Haimes  @ MIT         *
 *                                                                      *
 ************************************************************************
 */

#ifdef GEOM_EGADS

/*
 * Copyright (C) 2011  John F. Dannenhoffer, III (Syracuse University)
 *
 * This library is free software; you can redistribute it and/or
 *    modify it under the terms of the GNU Lesser General Public
 *    License as published by the Free Software Foundation; either
 *    version 2.1 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 *    but WITHOUT ANY WARRANTY; without even the implied warranty of
 *    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *    Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 *    License along with this library; if not, write to the Free Software
 *    Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
 *     MA  02110-1301  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#ifdef WIN32
    #define snprintf _snprintf
#endif

#include "egads.h"

/* ADDITIONAL ATTRIBUTE TYPE */
#define ATTRREALSEN 4

#define TWOPI 6.2831853071795862319959269

/* static variables used to hold info between function calls */
static char   *FileName = NULL;

/* number of Extract UDPs */
static int    numUdp    = 0;

/* arrays to hold info for each of the UDPs */
static ego    *ebodys   = NULL;
static int    *BodyNum  = NULL;
static int    **Quilts  = NULL;
static int    *Extend   = NULL;

/* routines defined below */


/*
 ************************************************************************
 *                                                                      *
 *   udpErrorStr - print error string                                   *
 *                                                                      *
 ************************************************************************
 */

static /*@null@*/ char *
udpErrorStr(int stat)
{
    char *string;

    string = EG_alloc(25*sizeof(char));
    if (string == NULL) {
        return string;
    }
    snprintf(string, 25, "EGADS status = %d", stat);

    return string;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpInitialize - initialize and get info about the list of arguments*
 *                                                                      *
 ************************************************************************
 */

int
udpInitialize(int    *nArgs,
              char   ***namex,
              int    **typex,
              int    **idefaulx,
              double **ddefaulx)
{
    int    *type, *idefault;
    char   **name;
    double *ddefault;

    /* make the arrays */
    ebodys  = (ego    *) EG_alloc(sizeof(ego    ));
    BodyNum = (int    *) EG_alloc(sizeof(int    ));
    Quilts  = (int   **) EG_alloc(sizeof(int    ));
    Extend  = (int    *) EG_alloc(sizeof(int    ));

    if (ebodys == NULL || BodyNum == NULL || Quilts == NULL || Extend == NULL) {
        return EGADS_MALLOC;
    }

    /* initialize the array elements that will hold the "current" settings */
    ebodys[0]  = NULL;
    BodyNum[0] = 0;
    Quilts[0]  = (int*) EG_alloc(6*sizeof(int));
    Extend[0]  = 0;

    if (Quilts[0] == NULL) {
        return EGADS_MALLOC;
    }

    /* set up returns that describe the UDP */
    *nArgs    = 4;
    *namex    = NULL;
    *typex    = NULL;
    *idefaulx = NULL;
    *ddefaulx = NULL;

    *namex = name = (char **) EG_alloc((*nArgs)*sizeof(char *));
    if (*namex == NULL) {
        return EGADS_MALLOC;
    }

    *typex = type = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*typex == NULL) {
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *idefaulx = idefault = (int *) EG_alloc((*nArgs)*sizeof(int));
    if (*idefaulx == NULL) {
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    *ddefaulx = ddefault = (double *) EG_alloc((*nArgs)*sizeof(double));
    if (*ddefaulx == NULL) {
        EG_free(*idefaulx);
        EG_free(*typex);
        EG_free(*namex);
        return EGADS_MALLOC;
    }

    name[0]     = "FileName";
    type[0]     = ATTRSTRING;
    idefault[0] = 0;
    ddefault[0] = 0.0;

    name[1]     = "BodyNum";
    type[1]     = ATTRINT;
    idefault[1] = 0;
    ddefault[1] = 0;

    name[2]     = "Quilts";
    type[2]     = ATTRSTRING;
    idefault[2] = 1;
    ddefault[2] = 0;

    name[3]     = "Extend";
    type[3]     = ATTRINT;
    idefault[3] = 0;
    ddefault[3] = 0;

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpReset - reset the arguments to their defaults                   *
 *                                                                      *
 ************************************************************************
 */

int
udpReset(int flag)
{
    int   iudp;

    if (FileName != NULL) {
        EG_free(FileName);
        FileName = NULL;
    }

    /* reset the "current" settings */
    if (flag == 0) {
        BodyNum[0]    = 0;
        Quilts[ 0][0] = 0;
        Quilts[ 0][1] = 0;
        Quilts[ 0][2] = 0;
        Quilts[ 0][3] = 0;
        Quilts[ 0][4] = 0;
        Quilts[ 0][5] = 0;
        Extend[ 0]    = 0;

    /* called when closing up */
    } else {
        for (iudp = 0; iudp <= numUdp; iudp++) {
            if (Quilts[iudp] != NULL) {
                EG_free(Quilts[iudp]);
                Quilts[iudp] = NULL;
            }
        }

        EG_free(BodyNum);  BodyNum = NULL;
        EG_free(ebodys );  ebodys  = NULL;
        EG_free(Quilts );  *Quilts = NULL;
        EG_free(Extend );  Extend  = NULL;

        numUdp = 0;
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSet - set an argument                                           *
 *                                                                      *
 ************************************************************************
 */

int
udpSet(char *name,
       char *value)
{
    int  i, len, iquilt, icount, jcount;
    char *pEnd, lowername[257], defn[128];

    if        (name  == NULL) {
        return EGADS_NONAME;
    } else if (strlen(name) == 0 || strlen(name) > 255) {
        return EGADS_NONAME;
    } else if (value == NULL) {
        return EGADS_NULLOBJ;
    }

    for ( i = 0; i <= strlen(name); i++) {
        lowername[i] = tolower(name[i]);
    }

    len = strlen(value);
    if (len == 0) {
        return EGADS_NODATA;
    }

    if        (strcmp(lowername, "filename") == 0) {
        if (FileName != NULL) {
            EG_free(FileName);
            FileName = NULL;
        }

        FileName = (char *) EG_alloc((len+1)*sizeof(char));
        if (FileName == NULL) {
            return EGADS_MALLOC;
        }

        strcpy(FileName, value);

    } else if (strcmp(lowername, "bodynum") == 0) {
        BodyNum[0] = strtol(value, &pEnd, 10);
        if (BodyNum[0] < 0) {
            printf("WARNING:: udpSet: BodyNum = %d < 0\n", BodyNum[0]);
            return EGADS_RANGERR;
        }

    } else if (strcmp(lowername, "quilts") == 0) {

        /* extract the data from the string */
        icount = 0;
        for (iquilt = 0; iquilt < 6; iquilt++) {
            jcount = 0;

            while (icount < strlen(value)) {
                if (value[icount] == ';') {
                    icount++;
                    break;
                } else {
                    defn[jcount  ] = value[icount];
                    defn[jcount+1] = '\0';
                    icount++;
                    jcount++;
                }
            }

            Quilts[0][iquilt] = strtol(defn, &pEnd, 10);
        }

    } else if (strcmp(lowername, "extend") == 0) {
        Extend[0] = strtol(value, &pEnd, 10);
        if (Extend[0] < 0) {
            printf("WARNING:: udpSet: Extend = %d < 0\n", Extend[0]);
            return EGADS_RANGERR;
        }

    } else {
        printf("WARNING:: udpSet: Parameter \"%s\" not known (should be filename, bodynum, quilts, or extend) \n", name);
        return EGADS_INDEXERR;

    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpExecute - execute the primitive                                 *
 *                                                                      *
 ************************************************************************
 */

int
udpExecute(ego  context,
           ego  *ebody,
           int  *nMesh,
           char **string)
{
    int     status = EGADS_SUCCESS;

    int     nbody, oclass, mtype, *senses, atype, alen, outLevel, ilist[6];
    int     inquilt, inode, iedge, iloop, iface, jface, copy, ichild, nchild;
    int     nface_extr, nface_temp, nshell_orig, nface_orig, nloop_orig, nedge_orig, nnode_orig;
    const int *tempIlist;
    double  data[18];
    ego     emodel, eref, ebody_orig, *ebodys_orig=NULL;
    ego     *eshells_orig=NULL, *eshells_copy=NULL;
    ego     *efaces_orig=NULL,  *efaces_copy=NULL, *efaces_extr=NULL, *efaces_temp=NULL;
    ego     *eloops_orig=NULL,  *eloops_copy=NULL;
    ego     *eedges_orig=NULL,  *eedges_copy=NULL;
    ego     *enodes_orig=NULL,  *enodes_copy=NULL;
    ego     *echilds_orig=NULL, *echilds_copy=NULL;

    /* default return values */
    *ebody  = NULL;
    *nMesh  = 0;
    *string = NULL;

    /* check if FileName was not given */
    if (FileName == NULL) {
        printf("WARNING:: udpExecute: no FileName given\n");
        return EGADS_EMPTY;
    }

    /* increment number of UDPs */
    numUdp++;

    /* increase the arrays to make room for the new UDP */
    ebodys  = (ego    *) EG_reall(ebodys,  (numUdp+1)*sizeof(ego    ));
    BodyNum = (int    *) EG_reall(BodyNum, (numUdp+1)*sizeof(int    ));
    Quilts  = (int   **) EG_reall(Quilts,  (numUdp+1)*sizeof(int    ));
    Extend  = (int    *) EG_reall(Extend,  (numUdp+1)*sizeof(int    ));

    if (ebodys == NULL || BodyNum == NULL || Quilts == NULL || Extend == NULL) {
        return EGADS_MALLOC;
    }

    /* store the udp information */
    ebodys[ numUdp] = NULL;
    BodyNum[numUdp] = BodyNum[0]; 
    Quilts[ numUdp] = (int*) EG_alloc(6*sizeof(int));
    Extend[ numUdp] = Extend[0];

    if (Quilts[numUdp] == NULL) {
        return EGADS_MALLOC;
    }

    Quilts[numUdp][0] = Quilts[0][0];
    Quilts[numUdp][1] = Quilts[0][1];
    Quilts[numUdp][2] = Quilts[0][2];
    Quilts[numUdp][3] = Quilts[0][3];
    Quilts[numUdp][4] = Quilts[0][4];
    Quilts[numUdp][5] = Quilts[0][5];

    /* load the Model from the file */
    status = EG_loadModel(context, 0, FileName, &emodel);
//$$$    printf("EG_loadModel() -> status=%d\n", status);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* find the identity of the requested Body */
    status = EG_getTopology(emodel, &eref, &oclass, &mtype, data,
                            &nbody, &ebodys_orig, &senses);
//$$$    printf("EG_getTopology(emodel) -> status=%d\n", status);
    if (status != EGADS_SUCCESS) goto cleanup;

    if (BodyNum[numUdp] >= nbody) {
        printf("WARNING:: udpExecute: BodyNum(=%d) >= nbody(=%d)\n", BodyNum[numUdp], nbody);
        status = EGADS_RANGERR;
        goto cleanup;
    }

    /* make list of Shells, Faces, Loops, Edges, and Nodes in ebody_orig */
    ebody_orig = ebodys_orig[BodyNum[numUdp]];

    status = EG_getBodyTopos(ebody_orig, NULL, SHELL, &nshell_orig, &eshells_orig);
//$$$    printf("EG_getBodyTopos(SHELL) -> status=%d, nshell_orig=%d\n", status, nshell_orig);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody_orig, NULL, FACE, &nface_orig, &efaces_orig);
//$$$    printf("EG_getBodyTopos(FACE) -> status=%d, nface_orig=%d\n", status, nface_orig);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody_orig, NULL, LOOP, &nloop_orig, &eloops_orig);
//$$$    printf("EG_getBodyTopos(LOOP) -> status=%d, nloop_orig=%d\n", status, nloop_orig);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody_orig, NULL, EDGE, &nedge_orig, &eedges_orig);
//$$$    printf("EG_getBodyTopos(EDGE) -> status=%d, nedge_orig=%d\n", status, nedge_orig);
    if (status != EGADS_SUCCESS) goto cleanup;

    status = EG_getBodyTopos(ebody_orig, NULL, NODE, &nnode_orig, &enodes_orig);
//$$$    printf("EG_getBodyTopos(NODE) -> status=%d, nnode_orig=%d\n", status, nnode_orig);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* for now, this only works with one Shell */
    if (nshell_orig != 1) {
        printf("WARNING:: udpExecute: wrong number of Shells\n");
        status = EGADS_RANGERR;
        goto cleanup;
    }

    /* allocate arrays for the copy */
    eshells_copy = EG_alloc(nshell_orig*sizeof(ego));
    efaces_copy  = EG_alloc(nface_orig *sizeof(ego));
    eloops_copy  = EG_alloc(nloop_orig *sizeof(ego));
    eedges_copy  = EG_alloc(nedge_orig *sizeof(ego));
    enodes_copy  = EG_alloc(nnode_orig *sizeof(ego));

    if (eshells_copy == NULL || efaces_copy == NULL || eloops_copy == NULL ||
        eedges_copy  == NULL || enodes_copy == NULL                          ) {
        return EGADS_MALLOC;
    }

    /* make list of the Faces to be extracted */
    nface_extr = 0;
    for (iface = 0; iface < nface_orig; iface++) {
        status = EG_attributeRet(efaces_orig[iface], "inquilt", &atype, &alen, &tempIlist, NULL, NULL);
        
        if (status == EGADS_SUCCESS) {
            inquilt = tempIlist[0];
        } else {
            inquilt = -1;
        }

        if (inquilt == Quilts[numUdp][0] ||
            inquilt == Quilts[numUdp][1] ||
            inquilt == Quilts[numUdp][2] ||
            inquilt == Quilts[numUdp][3] ||
            inquilt == Quilts[numUdp][4] ||
            inquilt == Quilts[numUdp][5]   ) {
            nface_extr++;
        }
    }

    efaces_extr = EG_alloc(nface_extr*sizeof(ego));

    nface_extr = 0;
    for (iface = 0; iface < nface_orig; iface++) {
        status = EG_attributeRet(efaces_orig[iface], "inquilt", &atype, &alen, &tempIlist, NULL, NULL);
        
        if (status == EGADS_SUCCESS) {
            inquilt = tempIlist[0];
        } else {
            inquilt = -1;
        }

        if (inquilt == Quilts[numUdp][0] ||
            inquilt == Quilts[numUdp][1] ||
            inquilt == Quilts[numUdp][2] ||
            inquilt == Quilts[numUdp][3] ||
            inquilt == Quilts[numUdp][4] ||
            inquilt == Quilts[numUdp][5]   ) {
            efaces_extr[nface_extr++] = efaces_orig[iface];
//$$$            printf("extract face %d (efaces_extr[%d]=%x\n",
//$$$                   iface, nface_extr-1, (int)efaces_extr[nface_extr-1]);
        }
    }

    /* copy the Nodes associated with one of the Faces to be extracted */
    for (inode = 0; inode < nnode_orig; inode++) {
        status = EG_getBodyTopos(ebody_orig, enodes_orig[inode], FACE, &nface_temp, &efaces_temp);
//$$$            printf("EG_getBodyTopos(FACE) -> status=%d, nface_temp=%d\n", status, nface_temp);
        if (status != EGADS_SUCCESS) goto cleanup;

        copy = 0;
        for (iface = 0; iface < nface_temp; iface++) {
            for (jface = 0; jface < nface_extr; jface++) {
                if (efaces_temp[iface] == efaces_extr[jface]) {
                    copy++;
                    break;
                }
            }
            if (copy > 0) break;
        }

        EG_free(efaces_temp);
        efaces_temp = NULL;

        if (copy > 0) {
            status = EG_copyObject(enodes_orig[inode], NULL, &(enodes_copy[inode]));
//$$$            printf("EG_copyObject(node[%d]) -> status=%d\n", inode, status);
            if (status != EGADS_SUCCESS) goto cleanup;
        } else {
//$$$                printf("not copying %d\n", inode);
            enodes_copy[inode] = NULL;
        }
    }

    /* copy the Edges associated with one of the Faces to be extracted */
    for (iedge = 0; iedge < nedge_orig; iedge++) {
        status = EG_getBodyTopos(ebody_orig, eedges_orig[iedge], FACE, &nface_temp, &efaces_temp);
//$$$            printf("EG_getBodyTopos(FACE) -> status=%d, nface_temp=%d\n", status, nface_temp);
        if (status != EGADS_SUCCESS) goto cleanup;
        
        copy = 0;
        for (iface = 0; iface < nface_temp; iface++) {
            for (jface = 0; jface < nface_extr; jface++) {
                if (efaces_temp[iface] == efaces_extr[jface]) {
                    copy++;
                    break;
                }
            }
            if (copy > 0) break;
        }
        
        EG_free(efaces_temp);
        efaces_temp = NULL;
        
        if (copy > 0) {
            status = EG_getTopology(eedges_orig[iedge], &eref, &oclass, &mtype, data,
                                    &nchild, &echilds_orig, &senses);
//$$$            printf("EG_getTopology(edge[%d]) => status=%d\n", iedge, status);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            echilds_copy = EG_alloc(nchild*sizeof(ego));
            for (ichild = 0; ichild < nchild; ichild++) {
                echilds_copy[ichild] = NULL;
            }
            
            for (inode = 0; inode < nnode_orig; inode++) {
                for (ichild = 0; ichild < nchild; ichild++) {
                    if (echilds_orig[ichild] == enodes_orig[inode]) {
                        echilds_copy[ichild]  = enodes_copy[inode];
                    }
                }
            }
            
            status = EG_makeTopology(context, eref, oclass, mtype, data,
                                     nchild, echilds_copy, senses, &(eedges_copy[iedge]));
//$$$            printf("EG_makeTopology(edge[%d]) -> status=%d\n", iedge, status);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            status = EG_attributeDup(eedges_orig[iedge], eedges_copy[iedge]);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            EG_free(echilds_copy);
            echilds_copy = NULL;
        } else {
            eedges_copy[iedge] = NULL;
        }
    }
    
    /* copy the Loops associated with one of the Faces to be extracted */
    for (iloop = 0; iloop < nloop_orig; iloop++) {
        status = EG_getBodyTopos(ebody_orig, eloops_orig[iloop], FACE, &nface_temp, &efaces_temp);
//$$$            printf("EG_getBodyTopos(FACE) -> status=%d, nface_temp=%d\n", status, nface_temp);
        if (status != EGADS_SUCCESS) goto cleanup;
        
        copy = 0;
        for (iface = 0; iface < nface_temp; iface++) {
            for (jface = 0; jface < nface_extr; jface++) {
                if (efaces_temp[iface] == efaces_extr[jface]) {
                    copy++;
                    break;
                }
            }
            if (copy > 0) break;
        }
        
        EG_free(efaces_temp);
        efaces_temp = NULL;
        
        if (copy > 0) {
            status = EG_getTopology(eloops_orig[iloop], &eref, &oclass, &mtype, data,
                                    &nchild, &echilds_orig, &senses);
//$$$            printf("EG_getTopology(Loop[%d]) => status=%d\n", iloop, status);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            echilds_copy = EG_alloc(2*nchild*sizeof(ego));
            for (ichild = 0; ichild < nchild; ichild++) {
                echilds_copy[ichild       ] = NULL;
                echilds_copy[ichild+nchild] = echilds_orig[ichild+nchild];
            }
            
            for (iedge = 0; iedge < nedge_orig; iedge++) {
                for (ichild = 0; ichild < nchild; ichild++) {
                    if (echilds_orig[ichild] == eedges_orig[iedge]) {
                        echilds_copy[ichild]  = eedges_copy[iedge];
                    }
                }
            }
            
            outLevel = EG_setOutLevel(context, 0);
            status = EG_makeTopology(context, eref, oclass, mtype, data,
                                     nchild, echilds_copy, senses, &(eloops_copy[iloop]));
            (void) EG_setOutLevel(context, outLevel);
//$$$            printf("EG_makeTopology(Loop[%d]) -> status=%d\n", iloop, status);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            status = EG_attributeDup(eloops_orig[iloop], eloops_copy[iloop]);
            if (status != EGADS_SUCCESS) goto cleanup;
            
            EG_free(echilds_copy);
            echilds_copy = NULL;
        } else {
            eloops_copy[iloop] = NULL;
        }
    }

    /* make the new extracted Faces */
    for (iface = 0; iface < nface_extr; iface++) {
        status = EG_getTopology(efaces_extr[iface], &eref, &oclass, &mtype, data,
                                &nchild, &echilds_orig, &senses);
//$$$        printf("EG_getTopology(Face[%d]) -> status=%d\n", iface, status);
        if (status != EGADS_SUCCESS) goto cleanup;

        echilds_copy = EG_alloc(nchild*sizeof(ego));

        for (ichild = 0; ichild < nchild; ichild++) {
            echilds_copy[ichild] = NULL;
        }

        for (iloop = 0; iloop < nloop_orig; iloop++) {
            for (ichild = 0; ichild < nchild; ichild++) {
                if (echilds_orig[ichild] == eloops_orig[iloop]) {
                    echilds_copy[ichild] =  eloops_copy[iloop];
                }
            }
        }

        status = EG_makeTopology(context, eref, oclass, mtype, data,
                                 nchild, echilds_copy, senses, &(efaces_copy[iface]));
//$$$        printf("EG_makeTopology(Face[%d]) -> status=%d\n", iface, status);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_attributeDup(efaces_extr[iface], efaces_copy[iface]);
        if (status != EGADS_SUCCESS) goto cleanup;

        status = EG_attributeRet(efaces_copy[iface], "inquilt", &atype, &alen, &tempIlist, NULL, NULL);
        if (status != EGADS_SUCCESS) goto cleanup;
//$$$        printf("iface=%d, tempIlist[0]=%d\n", iface, tempIlist[0]);

                                               ilist[0] = 0;
        if (tempIlist[0] == Quilts[numUdp][0]) ilist[0] = 1;
        if (tempIlist[0] == Quilts[numUdp][1]) ilist[0] = 2;
        if (tempIlist[0] == Quilts[numUdp][2]) ilist[0] = 3;
        if (tempIlist[0] == Quilts[numUdp][3]) ilist[0] = 4;
        if (tempIlist[0] == Quilts[numUdp][4]) ilist[0] = 5;
        if (tempIlist[0] == Quilts[numUdp][5]) ilist[0] = 6;
//$$$        printf("ilist[0]=%d\n", ilist[0]);

        status = EG_attributeAdd(efaces_copy[iface], "inquilt", ATTRINT, 1, ilist, NULL, NULL);
        if (status != EGADS_SUCCESS) goto cleanup;

        EG_free(echilds_copy);
        echilds_copy = NULL;
    }

    /* make an open Shell */
    echilds_copy = EG_alloc(nface_extr*sizeof(ego));

    nchild = 0;
    for (iface = 0; iface < nface_orig; iface++) {
        if (efaces_copy[iface] != NULL) {
            echilds_copy[nchild++] = efaces_copy[iface];
        }
    }

    status = EG_makeTopology(context, NULL, SHELL, OPEN, data,
                             nchild, echilds_copy, NULL, &eshells_copy[0]);
//$$$    printf("EG_makeTopology(Shell) -> status=%d\n", status);
    if (status != EGADS_SUCCESS) goto cleanup;

    EG_free(echilds_copy);
    echilds_copy = NULL;

    /* make the new body */
    status = EG_makeTopology(context, NULL, BODY, SHEETBODY, NULL,
                             1, eshells_copy, NULL, ebody);
//$$$    printf("EG_makeTopology(body) -> status=%d\n", status);
    if (status != EGADS_SUCCESS) goto cleanup;

    /* remove the original Model */
    outLevel = EG_setOutLevel(context, 0);
    status = EG_deleteObject(emodel);
    (void) EG_setOutLevel(context, outLevel);
//$$$    printf("EG_deleteObject(emodel) -> status=%d\n", status);
    if (status != EGADS_SUCCESS) goto cleanup;
            
cleanup:
    if (eshells_copy != NULL) EG_free(eshells_copy);
    if (efaces_copy  != NULL) EG_free(efaces_copy);
    if (eloops_copy  != NULL) EG_free(eloops_copy);
    if (eedges_copy  != NULL) EG_free(eedges_copy);
    if (enodes_copy  != NULL) EG_free(enodes_copy);

    if (eshells_orig != NULL) EG_free(eshells_orig);
    if (efaces_orig  != NULL) EG_free(efaces_orig);
    if (eloops_orig  != NULL) EG_free(eloops_orig);
    if (eedges_orig  != NULL) EG_free(eedges_orig);
    if (enodes_orig  != NULL) EG_free(enodes_orig);

    if (efaces_extr  != NULL) EG_free(efaces_extr);

    if (status != EGADS_SUCCESS) {
        *string = udpErrorStr(status);
    }

    return status;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpMesh - return mesh associated with primitive                    *
 *                                                                      *
 ************************************************************************
 */

int
udpMesh(ego    ebody,
        int    iMesh,
        int    *imax,
        int    *jmax,
        int    *kmax,
        double **mesh)
{
    int    iudp, judp;

    *imax = 0;
    *jmax = 0;
    *kmax = 0;
    *mesh = NULL;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (iMesh != 0) {
        printf("WARNING:: udpMesh: iMesh %d != 0\n", iMesh);
    }

    return EGADS_SUCCESS;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpSensitivity - return sensitivity derivatives for the "real" argument *
 *                                                                      *
 ************************************************************************
 */

int
udpSensitivity(ego    ebody,
               char   *vname,
  /*@unused@*/ int    npts,
  /*@unused@*/ int    *fIndices,
  /*@unused@*/ double *uvs,
  /*@unused@*/ double *dxdname)
{
    int iudp, judp;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}


/*
 ************************************************************************
 *                                                                      *
 *   udpStepSize - return sensitivity step size of finite differencing  *
 *                                                                      *
 ************************************************************************
 */

int
udpStepSize(ego    ebody,
            char   *vname,
            double *delta)
{
    int iudp, judp;

    *delta = 0.0;

    /* check that ebody matches one of the ebodys */
    iudp = 0;
    for (judp = 1; judp <= numUdp; judp++) {
        if (ebody == ebodys[judp]) {
            iudp = judp;
            break;
        }
    }
    if (iudp <= 0) {
        return EGADS_NOTMODEL;
    }

    if (vname  == NULL) {
        return EGADS_NONAME;
    }

    /* this routine is not written yet */
    return EGADS_NOLOAD;
}

#endif
